#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2018 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""PySceneDetect Unit Test Suite

To run all available tests run `pytest -v` from the parent directory
(i.e. the root project folder of PySceneDetect containing the scenedetect/
and tests/ folders).  This will automatically find and run all of the
test cases in the tests/ folder and display the results.
"""
